# Perform checks that can only be performed at the end

# Input: the global information environment
#
# Output: NULL
#
# Details: returns errors or warnings based on java/shiny requirements
#
finalChecks = function(globalInfo, checkDir) {
  # Check internal links
  links = getGlobal("internalLinks", globalInfo)
  if (!is.null(links)) {
    locs = getGlobal("internalLocs", globalInfo)
    if (is.null(locs)) {
      stop("@", links[1,"lineNum"], " \\shinyInternalLink, but app has no \\shinyInternalLoc's")
    }
    for (i in 1:nrow(links)) {
      if (is.na(match(links[i, "name"], locs))) {
        stop("@", links[i, "lineNum"], " \\shinyInternalLink to '", links[i, "name"], 
             "' has no matching \\shinyInternalLoc")
      }
    }
  } # end checking internal links

  # Check the conditional panel conditions are valid
  inputIds = getGlobal("inputIds", globalInfo)
  cps = getGlobal("conditionalPanels", globalInfo)
  if (!is.null(cps)) {
    for (cp in 1:nrow(cps)) {
      cId = cps[cp, "inputId"]
      pos = match(cId, inputIds$inputId)
      if (is.na(pos)) {
        stop("@", cps[cp, "lineNum"], " \\shinyConditionalPanel references missing ",
             "inputId='", cId, "'")
      }
      cType = inputIds[pos, "type"]
      ## more checks here !!!!!!!!!! ??????????????
      if (cType == "shinyDropdown") {
        if (!exists("dropdownChoices")) dropdownChoices = getGlobal("dropdownChoices", globalInfo)
        cChoicesList = dropdownChoices[[cId]]
        assure(!is.null(cChoicesList), 77926501)
        cChoices = unlist(gsub("[ ]*$", "", gsub("^[ ]*", "", 
                        strsplit(cChoicesList[["choices"]], ","))))
        if (is.na(match(cps[cp, "condition"], cChoices))) {
          stop("@", cps[cp, "lineNum"], " \\shinyConditionalPanel reference value '",
               cps[cp, "condition"], "', but that is not one of the dropdown choices on line ",
               inputIds[pos, "lineNum"])
        }
      }
    }
  }

  # Check that files are present
  filenames = getGlobal("filenames", globalInfo)
  if (!is.null(filenames)) {
    for (i in 1:nrow(filenames)) {
      if (!file.exists(file.path("www", filenames[i, "filename"]))) {
        warning("expecting ", filenames[i, "type"], " ", filenames[i, "filename"], 
                " in 'www' folder")
      }
    }
  }

  # Check feedbackForFoo() if FooCheck.txt exists  ## later add hiddenFeedback
  # FooCheckOut.txt is written with feedback text for each code block in FooCheck.txt.
  # Code blocks are separated by a line starting with at least three dashes ("---").
  # If FooCheckOutOK.txt exists, the current output is checked against the contents
  # of that file, and a warning is given if the two files do not match exactly.
  # Note: Testing of "carryOverVars=TRUE" is not implemented.
  #
  if (!is.null(checkDir)) {
    temp = file.info(checkDir)
    if (is.na(temp[1, "isdir"]) || !temp[1, "isdir"]) stop("checkDir '", checkDir, "' is not a directory")
  }
  feedbacks = getGlobal("RFeedback", globalInfo)
  hiddens = getGlobal("RHiddenFeedback", globalInfo)
  #
  if (!is.null(feedbacks)) {
    # For each RFeedback in the app
    for (fbNum in 1:nrow(feedbacks)) {
      fb = as.character(feedbacks[fbNum, "id"])
      hasHidden = !is.null(hiddens) && any(as.character(hiddens$id)==fb)
      fbR = paste0("feedbackFor", fb, ".R")
      temp = try(source(fbR), silent=TRUE)
      if (is(temp, "try-error")) stop("syntax error in ", fbR, ".  Check your RFeedback code.")
      if (hasHidden) {
        fbHR = paste0("hiddenFeedbackFor", fb, ".R")
        temp = try(source(fbHR), silent=TRUE)
        if (is(temp, "try-error")) stop("syntax error in ", fbHR, ".  Check your RHiddenFeedback code.")
      }
      inFile = paste0(fb, "Check.txt")
      usingCheckDir = FALSE
      if (!file.exists(inFile) && !is.null(checkDir)) {
        usingCheckDir = TRUE
        inFile = file.path(checkDir, inFile)
      }
      if (file.exists(inFile)) {
        inCode = try(readLines(inFile))
        if (is(inCode, "try-error")) stop(inFile, " exists, but is not a readable text file")
        inRLE = rle(substring(inCode, 1, 3) != "---")
        inBlocks = which(inRLE$values == TRUE)
        if (length(inBlocks)>0) {
          outFile = paste0(fb, "CheckOut.txt")
          if (usingCheckDir) outFile = file.path(checkDir, outFile)
          write(paste("#", fbR), file=outFile)
          inEnd = cumsum(inRLE$lengths)
          inStart = c(1, inEnd+1)
          inCode = lapply(inBlocks, function(x) inCode[inStart[x]:inEnd[x]])
          for (code in inCode) {
            write(code, file=outFile, append=TRUE)
            write("---", file=outFile, append=TRUE)
            checkEnv = feedbacks[fbNum, "vars"]
            if (checkEnv=="") {
              checkEnv = list2env(list()) 
            } else {
              expr = try(parse(text=paste0("list(", checkEnv, ")")))
              if (is(expr, "try-error")) stop("unexpected error #1 in finalChecks()")
              checkList = try(eval(expr))
              if (is(expr, "try-error")) stop("unexpected error #2 in finalChecks()")
              checkEnv = list2env(checkList)
            }
            cdr = runR(code, env=checkEnv, allowSemicolons=feedbacks[fbNum, "allowSemicolons"])
            vars = eapply(checkEnv, identity)
            
            expr = parse(text=paste0("null = feedbackFor", fb, 
                                     "(unlist(cdr$input), cdr$input, unlist(cdr$output), cdr$output, ",
                                     "cdr$nefarious, length(cdr$input)==0, cdr$errorMessage, length(cdr$errorMessage)>0, ",
                                     "cdr$errorCode, vars, names(vars), suppressLogging=TRUE)"))
            rslt = capture.output(eval(expr))
            write(rslt, file=outFile, append=TRUE)
            if (hasHidden) {
              write("((((((((((((((((((((((((((((((((((((", file=outFile, append=TRUE)
              expr = parse(text=paste0("null = hiddenFeedbackFor", fb, 
                                       "(unlist(cdr$input), cdr$input, unlist(cdr$output), cdr$output, ",
                                       "cdr$nefarious, length(cdr$input)==0, cdr$errorMessage, length(cdr$errorMessage)>0, ",
                                       "cdr$errorCode, vars, names(vars), suppressLogging=TRUE)"))
              rslt = capture.output(eval(expr))
              write(rslt, file=outFile, append=TRUE)
              write("))))))))))))))))))))))))))))))))))))", file=outFile, append=TRUE)
            }
            write("------------------------------------", file=outFile, append=TRUE)
          } # end for each individual test of the feedback
          #
          # Check results against "OK" results
          okFile = paste0(fb, "CheckOutOK.txt")
          if (usingCheckDir) okFile = file.path(checkDir, okFile)
          hasOK = file.exists(okFile)
          if (hasOK) {
            okText = try(readLines(okFile))
            if (is(okText, "try-error")) stop("cannot read ", okFile)
            currText = readLines(outFile)
            if (length(okText) != length(currText) || !all(okText==currText))
              warning("current feedback does not match ", okFile)
          }
        } # end if blocks of code exist in the "check" file
      } # end if "check" input exists
    } # end for each feedback block
  } # end if there are any feedbacks

  invisible(NULL)
}
